/* Part3 : About Ring Topology 
 * Consider a set of processes arranged in a ring.
 * Each processor stores its rank in MPI_COMM_WORLD in an integer and
 * sends this value onto the processor on its right.
 * The processors continue passing on the values they receive 
 * until they get their own rank back. Each process should finish
 * by printing out the sum of the values.
 *
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/* The Definition of MPI */
#include "mpi.h"

#define PARTICLES 12000

extern void srand48();
extern double drand48();

void main( int argc, char **argv )
{
	int predecessor, successor, tag = 0;
	int my_rank, num_processes;
	int root = 0;

	double x[PARTICLES], y[PARTICLES];
	double correlation[PARTICLES];
	double total_correlation = 0.0;
    double sub_correlation = 0.0;

	int  local_p;

	double f(double a, double b);
	
	MPI_Status status;

	/* start up MPI	*/
	MPI_Init( &argc, &argv );

	/* get process rank */
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes */
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	local_p = PARTICLES/num_processes;

	/* Generate the initial values */
	srand48(my_rank * 117);
	for(int i = 0; i < local_p; i++) {
		x[i] = drand48();
		y[i] = drand48();
		correlation[i] = 0.0;
	}

	predecessor = ( my_rank + 1 ) % num_processes;
	successor = ( my_rank - 1 + num_processes ) % num_processes;

	for(int k = 0; k < num_processes;k++)
	{
		for(int j = 0; j < local_p; j++)
		  for(int i = 0; i < local_p; i++)
			correlation[i] += f( x[i], y[j] );
		
		MPI_Sendrecv_replace(y, local_p, MPI_DOUBLE, successor, tag, predecessor, tag, MPI_COMM_WORLD, &status);

	}

	for(int i = 0; i < local_p; i++)
		sub_correlation += correlation[i];

	if ( my_rank == root) {  
		total_correlation = sub_correlation;
			
        for( int i = 1; i < num_processes; i++ )
        {
            MPI_Recv( &sub_correlation, 1, MPI_DOUBLE, i, tag, MPI_COMM_WORLD, &status );
			total_coorelation += sub_correlation;
        }
	printf("Total Correlation is %f\n", total_correlation);

  	}
	else 
		MPI_Send(&sub_correlation, 1, MPI_DOUBLE, root, tag, MPI_COMM_WORLD );
		

	/* close MPI*/
	MPI_Finalize();
}


double f(double a, double b)
{
	return (abs(a*a + b*b));
}

